/**
  * A really basic 2D point class.
  * @author ric Plante
  */

package ep.cg;

import java.util.Vector;

public class Point {

  public static final Point ORIGIN = new Point(0,0);

  protected double x_;
  protected double y_;
  protected int ident_;
  public final static Point UNDEF = new Point(Double.NEGATIVE_INFINITY,
					      Double.NEGATIVE_INFINITY);

  /** This is rather special-purpose; indicates that the chain
    * starting at this vertex is part of the poly intersection.
    */
  private boolean fwIsXchain_ = false;
  /** This is rather special-purpose; it indicates that the chain
    * ending at this vertex is part of the poly intersection.
    */
  private boolean rvIsXchain_ = false;


  /**
    * The set of segments which <em>originate</em> from this point.
    * Note that a given sgment is not supposed to originate from two
    * different points, even though this is not enforced.
    */
  protected Vector segments_;

  public Point() {
    x_ = Double.NEGATIVE_INFINITY;
    y_ = Double.NEGATIVE_INFINITY;
    segments_ = new Vector();
  }

  public Point( double x, double y ) {
    x_ = x;
    y_ = y;
    segments_ = new Vector();
  }

  public Point( double x, double y, Segment pSegment ) {
    x_ = x;
    y_ = y;
    segments_ = new Vector(1);
    segments_.addElement(pSegment);
  }

  public void setIdent(int i) {
    ident_ = i;
  }

  public final double x() {
    return x_;
  }

  public final double y() {
    return y_;
  }

  public final void x(double x) {
    x_ = x;
  }

  public final void y(double y) {
    y_ = y;
  }

  public final void addSeg(Segment fSegment) {
    segments_.addElement(fSegment);
  }

  /**
    * Returns one of the segments originating from this point.
    * @arg segmentNumber The segment number, ranging 0..nbSegments()-1.
    */
  public final Segment getSeg(int segmentNumber) {
    return (Segment)(segments_.elementAt(segmentNumber));
  }

  public final void remSeg(Segment fSegment) {
    segments_.removeElement(fSegment);
  }

  public final void remSeg(Point p2) {
    int i;
    for (i = 0; i < segments_.size(); i++) {
      Segment s2 = (Segment)segments_.elementAt(i);
      if ( s2.getP1() == this && s2.getP2() == p2 ) {
	segments_.removeElement(s2);
      }
    }
  }
    
  public final void remSeg(int segmentNumber) {
    segments_.removeElementAt(segmentNumber);
  }

  public final void remSeg(Point p2, int ident) {
    Segment s = new Segment(this, p2, ident);
    int i;
    for (i = 0; i < segments_.size(); i++) {
      Segment s2 = (Segment)segments_.elementAt(i);
      if ( s.equals(s2) ) {
	segments_.removeElement(s2);
      }
    }
  }
  
  public final void clrSegs() {
    segments_.removeAllElements();
  }


  public final void fwIsXchain(boolean b) {
    fwIsXchain_ = b;
  }

  public final void rvIsXchain(boolean b) {
    rvIsXchain_ = b;
  }

  public final boolean fwIsXchain() {
    return fwIsXchain_;
  }

  public final boolean rvIsXchain() {
    return rvIsXchain_;
  }

  /**
    * Returns the identity of the point. Can be used to assign a
    * color...
    */
  public final int getIdent() {
    return ident_;
  }
  
  /**
    * Returns the number of segments which <em>originate</em> from
    * this point.
    */
  public final int nbSegs() {
    return segments_.size();
  }
    

  public final boolean equals( Point p ) {
    return ( (p.x_ == x_) && (p.y_ == y_) );
  }

  public double dist( Point p ) {
    double tmpx, tmpy;
    tmpx = ( x_ - p.x() );
    tmpx *= tmpx;
    tmpy = ( y_ - p.y() );
    tmpy *= tmpy;

    tmpx += tmpy;
    return Math.sqrt(tmpx);
  }

  public void add( Point p ) {
    x_ += p.x();
    y_ += p.y();
  }

  public void minus( Point p ) {
    x_ -= p.x();
    y_ -= p.y();
  }

  /**
    * Substracts point p2 from p1 and returns a new point.
    */
  public final static Point minus( Point p1, Point p2 ) {
    return new Point( p1.x_ - p2.x_, p1.y_ - p2.y_ );
  }

  /**
    * Scale the point by s about the origin.
    */
  public final void scale( double s ) {
    x_ *= s;
    y_ *= s;
  }

  /**
    * Consider the point as a vector and normalize it.
    */
  public final void normalize() {
    double tmp;
    tmp = Math.sqrt( x_*x_ + y_*y_ );
    x_ /= tmp;
    y_ /= tmp; 
  }

  public String toString() {

    int i, n = segments_.size();

    StringBuffer outs = new StringBuffer("Point ("+x_+", "+y_+")\n");

    for (i=0; i<n; i++)
      outs.append("  seg: "+
			 ((Segment)segments_.elementAt(i)).toString() +
		  "\n");
    return outs.toString();
  }
}
