package ep.cg;

import java.util.Vector;

/**
  * A basic polygon class which uses our homemade Point class, and
  * therefore is based on floating point math rather than integer math like
  * the polygon class in awt.
  * Points are forced to be in clockwise order.
  *
  * @author ric Plante
  **/

public class Polygon {

  protected Vector pts_;
  private int ident_;


  public Polygon() {
    pts_ = new Vector();
  }

  public void setIdent(int i) {
    ident_ = i;
  }

  public int getIdent() {
    return ident_;
  }

  /**
    * This method will throw a NonConvexPolygonException if the point added
    * would make the polygon intersect itself.
    * The point will be silently ignored if it is colinear with the two
    * last points added, or with the first and last points added.
    * @throws ColinearPointsException
    * @throws NonConvexPolygonException
    */

  public void addPoint( Point p ) throws NonConvexPolygonException {
    Point l1,l2;
    
    if ( ! checkNewConvex(p) )
      throw new NonConvexPolygonException();
    else    
      pts_.addElement( p );
  }

  /**
    * This method can be used to add a point without any convexity
    * check. Use this only when you're _sure_ it will be convex,
    * such as when intersecting two convex polygons...
    */
  public void addPointNoCheck( Point p) {
    pts_.addElement( p );
  }
  
  /**
    * Deletes a point. Fails silently if the point does not exist.
    * The parameter passed should be a reference to the point to delete.
    */
  public void delPoint( Point p ) {
    pts_.removeElement( p );
  }

  /**
    * This method checks if adding p to the polygon would keep it
    * convex.
    */
  public boolean checkNewConvex(double x, double y) {
    return checkNewConvex( new Point(x,y) );
  }


  public boolean checkNewConvex(Point p) {
    int size = pts_.size();
    
    if (size < 2)
      return true;

    Point last     = (Point)pts_.elementAt(size-1);
    Point first    = (Point)pts_.elementAt(0);

    if ( Side.whichSide(last, first, p) != Side.UNDER ) {
      return false;
    }

    if ( size > 2) {
      
      Point prevlast = (Point)pts_.elementAt(size-2);
      Point second   = (Point)pts_.elementAt(1);

      if ( Side.whichSide(prevlast,last,p) != Side.OVER ||
	   Side.whichSide(first,second,p)  != Side.OVER ) {
	return false;
      }
    }

    return true;
  }


  /***** DEBUG VERSION to check angles ******/
  /***
  public boolean checkConvex(Point p) {
    int idx = pts_.indexOf(p);
    int siz = pts_.size();

    int previdx, nextidx;

    if (idx < 0) {
      return false;
    }

    if (siz < 2)
      return true;

    if (idx == 0)
      previdx = siz - 1;
    else
      previdx = idx - 1;

    if (idx == siz -1)
      nextidx = 0;
    else
      nextidx = idx + 1;

    Point prev = (Point)pts_.elementAt(previdx);
    Point next = (Point)pts_.elementAt(nextidx);

    if ( siz > 2) {

      int prevprevidx = 0, nextnextidx = 0;
      
      
      if (previdx == 0)
	prevprevidx = siz - 1;
      else
	prevprevidx = previdx - 1;
      
      if (nextidx == siz - 1)
	nextnextidx = 0;
      else
	nextnextidx = nextidx + 1;

      Point prevprev = (Point)pts_.elementAt(prevprevidx);
      Point nextnext = (Point)pts_.elementAt(nextnextidx);


      Point pold = new Point(prev.x(),prev.y());
      pold.minus(prevprev);
      pold.normalize();
      Point pnew = new Point(p.x(), p.y());
      pnew.minus(prev);
      pnew.normalize();
      
      double a = pold.x()*pnew.x() + pold.y()*pnew.y();
      System.out.println("Angle is "+ ((Math.acos(a)/(2*Math.PI))*360) );
      

    }

    return true;
  }
  ***/

  public boolean checkConvex(Point p) {
    int idx = pts_.indexOf(p);
    int siz = pts_.size();

    int previdx, nextidx;
    
    if (idx < 0) {
      return false;
    }

    if (siz < 2)
      return true;

    if (idx == 0)
      previdx = siz - 1;
    else
      previdx = idx - 1;

    if (idx == siz -1)
      nextidx = 0;
    else
      nextidx = idx + 1;

    Point prev = (Point)pts_.elementAt(previdx);
    Point next = (Point)pts_.elementAt(nextidx);

    if ( Side.whichSide(prev, next, p) != Side.UNDER ) {
      return false;
    }

    if ( siz > 2) {

      int prevprevidx = 0, nextnextidx = 0;
      
      
      if (previdx == 0)
	prevprevidx = siz - 1;
      else
	prevprevidx = previdx - 1;
      
      if (nextidx == siz - 1)
	nextnextidx = 0;
      else
	nextnextidx = nextidx + 1;

      Point prevprev = (Point)pts_.elementAt(prevprevidx);
      Point nextnext = (Point)pts_.elementAt(nextnextidx);

      if ( Side.whichSide(prevprev,prev,p) != Side.OVER ||
	   Side.whichSide(next,nextnext,p)  != Side.OVER ) {
	return false;
      }
    }

    return true;
  }


  /**
    * Returns the point i of the polygon, where i e [0,n-1]
    */
  public Point getPoint(int i) {
    return (Point)pts_.elementAt(i);
  }

  /**
    * Returns number of points in poly.
    */
  public int size() {
    return pts_.size();
  }

  public String toString() {
    StringBuffer buf = new StringBuffer("Polygon: \n");
    int i, psize = pts_.size();

    for (i=0; i < psize; i++) {
      Point p = (Point)pts_.elementAt(i);
      buf.append(" Point: x=");
      buf.append(p.x());
      buf.append(" y=");
      buf.append(p.y());
      buf.append("\n");
    }
    return buf.toString();
    
  }
}
